---@type Tinkr
local Tinkr = ...

local startPrecision = GetTimePreciseSec()

---@class Bastion.Globals.SpellName : { [spellId]: string }

---@class Bastion
local Bastion = {
    CombatEvents = {},
    Enabled = false,
    Interval = 0.10,
    Created = GetTime(),
    Loaded = 0,
    NextUpdate = 0,
    TimeSinceLastUpdate = 0,
    TimeSinceLastOMUpdate = 0,
    TimeSinceLastMissleManagerUpdate = 0,
    TimeSinceLastTTDUpdate = 0,
    TimeSinceLastCombatTimerUpdate = 0,
    TimeSinceLastModuleUpdate = 0,
    modulesTicking = false,
    Updating = false,
    LastStart = 0,
    LastStop = 0,
    UpdateDuration = 0,
    Globals = {
        ---@type Bastion.Globals.SpellName
        SpellName = {}
    },
    ---@class Bastion.LoadedFiles.Table
    ---@field [number] { originalPath: string, loadedPath: string, reloadable: boolean, order: number, newPath: string }
    LoadedFiles = {},
    LoadThird = true,
    Paths = {
        Scripts = "scripts",
        BastionBase = "scripts/bastion",
        BastionScripts = "scripts/bastion/scripts",
        ThirdParty = "scripts/BastionScripts",
    },
    Log = function(message)
        Log(string.format("%s [%.2f] [Bastion]: %s", date("%H:%M:%S"), GetTimePreciseSec() - startPrecision, message))
    end,
}

--[[ function Bastion:__index(key)
    if Bastion[key] then
        return Bastion[key]
    end
    return rawget(self, key)
end ]]

local bastionFiles = {
    "~/src/ClassMagic/ClassMagic",
    "~/src/APL/APL",
    "~/src/APLActor/APLActor",
    "~/src/APLTrait/APLTrait",
    "~/src/Aura/Aura",
    "~/src/AuraTable/AuraTable",
    "~/src/Cache/Cache",
    "~/src/Cacheable/Cacheable",
    "~/src/Class/Class",
    "~/src/Command/Command",
    "~/src/Config/Config",
    "~/src/EventManager/EventManager",
    "~/src/Item/Item",
    "~/src/ItemBook/ItemBook",
    "~/src/Library/Library",
    "~/src/List/List",
    "~/src/Missile/Missile",
    "~/src/MissileManager/MissileManager",
    "~/src/Module/Module",
    "~/src/MythicPlusUtils/MythicPlusUtils",
    "~/src/Notification/Notification",
    "~/src/NotificationList/NotificationList",
    "~/src/ObjectManager/ObjectManager",
    "~/src/Refreshable/Refreshable",
    "~/src/Sequencer/Sequencer",
    "~/src/Spell/Spell",
    "~/src/SpellBook/SpellBook",
    "~/src/Timer/Timer",
    "~/src/TimeToDie/TimeToDie",
    "~/src/Unit/Unit",
    "~/src/UnitManager/UnitManager",
    "~/src/Util/Util",
    "~/src/Vector3/Vector3",
}

---@param filePath string
function Bastion:CheckIfLoaded(filePath)
    for i, file in ipairs(Bastion.LoadedFiles) do
        if file.loadedPath == filePath or file.originalPath == filePath or file.newPath == filePath then
            return true
        end
    end
    return false
end

---@param path string
---@param extension string
---@return string
local function AppendExtension(path, extension)
    return string.format("%s.%s", path, extension)
end

---@param path string
---@param extensions string|string[]
local function CheckFileExtensions(path, extensions)
    local exts = {}
    if type(extensions) == "string" then
        exts = { extensions }
    else
        exts = extensions
    end

    for i, extension in ipairs(exts) do
        local newPath = path
        if newPath:sub(extension:len() * -1) ~= extension then
            newPath = AppendExtension(newPath, extension)
        end
        if FileExists(newPath) then
            return newPath:sub(1, (extension:len() + 2) * -1), true
        end
    end
    return path, false
end

--- 0 = Failed, 1 = Success, 2 = Already Loaded
---@param filePath string | { filePath: string, reloadable: boolean }
---@param ... any
---@return 0|1|2, table
function Bastion:Require(filePath, ...)
    local loadedFile = {
        originalPath = filePath.filePath or filePath,
        newPath = filePath.filePath or filePath,
        loadedPath = filePath.filePath or filePath,
        reloadable = not filePath.reloadable and false or false,
        order = #Bastion.LoadedFiles + 1,
    }

    local pathResolutionShortcut = loadedFile.originalPath:sub(1, 1) == "@" and Bastion.Paths.BastionScripts or
        loadedFile.originalPath:sub(1, 1) == "~" and Bastion.Paths.BastionBase or
        loadedFile.originalPath:sub(1, 1) == "!" and Bastion.Paths.Scripts

    if pathResolutionShortcut then
        loadedFile.newPath = string.format("%s%s", pathResolutionShortcut, loadedFile.originalPath:sub(2))
        loadedFile.loadedPath = loadedFile.newPath
    end

    if DirectoryExists(loadedFile.newPath) then
        local fileList = ListFiles(loadedFile.newPath)
        if #fileList > 0 then
            local loadResults = {}
            for i = 1, #fileList do
                local file = loadedFile.newPath .. fileList[i]
                local status, returns = Bastion:Require(file, ...)
                table.insert(loadResults, { file = file, status = status, returns = returns })
            end
            return 1, loadResults
        end
        --Log(string.format("Bastion:Require - No files found in directory: %s", loadedFile.newPath))
        return 0, SafePack(nil)
    end

    local found = false
    -- Check if file path has a .lua or .luac extension. If not, try to add one and check if the file exists
    loadedFile.loadedPath, found = CheckFileExtensions(loadedFile.newPath, { "lua", "luac" })

    if not found then
        --Log(string.format("Bastion:Require - Not Found: %s (%s)", loadedFile.newPath, loadedFile.originalPath))
        return 0, SafePack(nil)
    end

    if not loadedFile.reloadable then
        if Bastion:CheckIfLoaded(loadedFile.loadedPath) then
            --Log(string.format("Bastion:Require - Already loaded: %s (%s)", loadedFile.newPath, loadedFile.originalPath))
            return 2, SafePack(nil)
        end
    end

    table.insert(Bastion.LoadedFiles, loadedFile)
    return 1, SafePack(require(loadedFile.loadedPath, Bastion, ...))
end

---@param load? boolean
---@param folder? string
local function LoadThird(load, folder)
    load = type(load) == "nil" and Bastion.LoadThird or load
    folder = folder or Bastion.Paths.ThirdParty
    if not load then
        return
    end
    local thirdPartyModulesFolders = ListFolders(folder)
    for i = 1, #thirdPartyModulesFolders do
        local moduleFolder = thirdPartyModulesFolders[i]
        local currentFolderDir = string.format("%s/%s", folder, moduleFolder)
        local loaderFilePath = string.format("%s/%s", currentFolderDir, "loader")
        if FileExists(loaderFilePath .. ".lua") or FileExists(loaderFilePath .. ".luac") then
            Bastion:Require(loaderFilePath, currentFolderDir)
        end
    end
end

---@param toggle? boolean
function Bastion:Toggle(toggle)
    Bastion.Enabled = type(toggle) ~= "nil" and toggle or not Bastion.Enabled
end

---@param interval number
function Bastion:SetInterval(interval)
    self.Interval = type(interval) == "number" and interval > 0 and interval or self.Interval
    --[[ Bastion.Ticker:Cancel()
    Bastion.Ticker = C_Timer.NewTicker(Bastion.Interval, function()
        Bastion.TickFunction(self)
    end) ]]
end

function Bastion:GetInterval()
    return Bastion.Interval
end

local combatEventSuffix = {
    ["_DAMAGE"] = true,
    ["_MISSED"] = true,
    ["_ABSORBED"] = true,
    ["_DRAIN"] = true,
    ["_LEECH"] = true,
    ["_INTERRUPT"] = true,
    ["_DISPEL"] = true,
    ["_DISPEL_FAILED"] = true,
    ["_STOLEN"] = true,
    ["_EXTRA_ATTACKS"] = true,
    ["_AURA_APPLIED"] = true,
    ["_AURA_REMOVED"] = true,
    ["_AURA_APPLIED_DOSE"] = true,
    ["_AURA_REMOVED_DOSE"] = true,
    ["_AURA_REFRESH"] = true,
    ["_AURA_BROKEN"] = true,
    ["_AURA_BROKEN_SPELL"] = true,
    ["_CAST_START"] = true,
    ["_CAST_SUCCESS"] = true,
    ["_CAST_FAILED"] = true,
    ["_INSTAKILL"] = true,
    ["_DURABILITY_DAMAGE"] = true,
    ["_DURABILITY_DAMAGE_ALL"] = true,
    ["_CREATE"] = true,
    ["_SUMMON"] = true,
    ["_RESURRECT"] = true,
    ["_EMPOWER_START"] = true,
    ["_EMPOWER_END"] = true,
    ["_EMPOWER_INTERRUPT"] = true,
}

local combatEventPrefix = {
    "SWING",
    "RANGE",
    "SPELL",
    "SPELL_PERIODIC",
    "SPELL_BUILDING",
    "ENVIRONMENTAL"
}

---@type Bastion.Module[]
local MODULES = {}

function Bastion:ModulesTicking()
    return self.modulesTicking
end

---@param force? boolean
function Bastion:TickModules(force)
    if not self.modulesTicking and (self.Enabled or force) then
        self.modulesTicking = true
        for i = 1, #MODULES do
            MODULES[i]:Tick()
        end
        self.modulesTicking = false
    end
end

-- Find a module by name
---@param name string
function Bastion:FindModule(name)
    for i = 1, #MODULES do
        if MODULES[i].name == name then
            return MODULES[i]
        end
    end
    return false
end

---@param module Bastion.Module
function Bastion:Register(module)
    local foundModule = Bastion:FindModule(module.name)
    if not foundModule then
        table.insert(MODULES, module)
        Bastion.Util:Print("Registered", module)
    else
        Bastion.Util:Print("Module already registered", module)
        return
    end
end

local loaded = false
function Bastion:Load()
    if loaded then
        return self
    end
    for suffix, _ in pairs(combatEventSuffix) do
        for _, prefix in pairs(combatEventPrefix) do
            Bastion.CombatEvents[prefix .. suffix] = true
        end
    end

    for i = 1, #bastionFiles do
        self:Require(bastionFiles[i])
    end

    ---#region Init Start
    Bastion.Globals.Command = Bastion.Command:New("bastion")
    Bastion.Globals.EventManager = Bastion.EventManager:New()
    Bastion.Globals.ItemBook = Bastion.ItemBook:New()
    Bastion.Globals.MissileManager = Bastion.MissileManager:New()
    Bastion.TimeToDie:Init()

    ---@param args { [2]: string }
    Bastion.Globals.Command:Register("module", "Toggle a module on/off", function(args)
        local module = Bastion:FindModule(args[2])
        if module then
            module:Toggle()
            if module.enabled then
                Bastion.Util:Print("Enabled", module.name)
            else
                Bastion.Util:Print("Disabled", module.name)
            end
        else
            Bastion.Util:Print("Module not found")
        end
    end)

    Bastion.Globals.MythicPlusUtils = Bastion.MythicPlusUtils:New()
    Bastion.Globals.NotificationList = Bastion.NotificationList:New()
    Bastion.ObjectManager = Bastion.ObjectManager:New()
    Bastion.Globals.SpellBook = Bastion.SpellBook:New(true)
    Bastion.Globals.UnitManager = Bastion.UnitManager:New()
    Bastion.Globals.Command:Register("toggle", "Toggle bastion on/off", function()
        Bastion.Util:TogglePrint()
    end)
    Bastion.Globals.Command:Register("debug", "Toggle debug mode on/off", function()
        Bastion.Util:ToggleDebug()
    end)
    Bastion.Globals.EventManager:RegisterWoWEvent("PLAYER_ENTERING_WORLD", function()
        Bastion.Globals.UnitManager:ResetObjects()
    end)
    Bastion.Globals.Command:Register('toggle', 'Toggle bastion on/off', function()
        self:Toggle()
        if self.Enabled then
            self.Util:Print("Enabled")
        else
            self.Util:Print("Disabled")
        end
    end)

    ---#endregion

    Bastion.Globals.CombatTimer = self.Timer:New("combat", function()
        return UnitAffectingCombat("player")
    end)

    ---@param unit UnitToken
    ---@param auras UnitAuraUpdateInfo
    Bastion.Globals.EventManager:RegisterWoWEvent("UNIT_AURA", function(unit, auras)
        local u = Bastion.Globals.UnitManager:Get(unit)

        if u then
            u:GetAuras():OnUpdate(auras)
        end
    end)

    Bastion.Globals.EventManager:RegisterWoWEvent("UNIT_SPELLCAST_SUCCEEDED", function(...)
        ---@type UnitIds, string, spellId
        local unit, castGUID, spellID = ...
        local spell = self.Globals.SpellBook:GetIfRegistered(spellID)

        if unit == "player" and spell then
            spell.lastCastAt = GetTime()

            if spell:GetPostCastFunction() then
                spell:GetPostCastFunction()(spell)
            end
        end
    end)

    local missed = {}

    Bastion.Globals.EventManager:RegisterWoWEvent("COMBAT_LOG_EVENT_UNFILTERED", function()
        local args = { CombatLogGetCurrentEventInfo() }
        local currTime = GetTime()
        ---@type string
        local subEvent = args[2]
        ---@type string
        local sourceGUID = args[4]
        ---@type string
        local destGUID = args[8]
        ---@type number
        local spellID = args[12]
        ---@type string
        local spellName = args[13]

        if subEvent == "SPELL_CAST_SUCCESS" then
            if not Bastion.Globals.SpellName[spellID] then
                Bastion.Globals.SpellName[spellID] = spellName
            end
        end

        local sourceUnit = Bastion.Globals.UnitManager:GetObject(sourceGUID)
        if sourceUnit and sourceUnit:Exists() then
            sourceUnit:SetLastCombatTime(currTime)
        end

        local destUnit = Bastion.Globals.UnitManager:GetObject(destGUID)

        if destUnit and destUnit:Exists() then
            destUnit:SetLastCombatTime(currTime)
        end

        if subEvent == "UNIT_DIED" or subEvent == "UNIT_DESTROYED" or subEvent == "PARTY_KILL" or subEvent == "SPELL_INSTAKILL" then
            Bastion.TimeToDie:UNIT_DIED(destGUID)
        end


        --[[ if sourceUnit and sourceUnit:IsValid() and sourceUnit:IsAffectingCombat() then
            sourceUnit:SetLastCombatTime(currTime)
        end

        local updateDestUnit = destUnit and destUnit:IsAffectingCombat()

        if Bastion.CombatEvents[subEvent] and not updateDestUnit or ((sourceUnit and destUnit) and not UnitThreatSituation(sourceUnit:GetOMToken(), destUnit:GetOMToken())) then
            updateDestUnit = true
        end

        if destUnit and updateDestUnit then
            destUnit:SetLastCombatTime(currTime)
        end ]]
    end)
    --[[ Bastion.Ticker = C_Timer.NewTicker(Bastion.Interval, function(this)
        Bastion.TickFunction(self)
    end) ]]

    --[[
        local currTime = GetTime()
        if currTime >= self.NextUpdate then
            self.NextUpdate = currTime + self.Interval
            if not self.Globals.CombatTimer:IsRunning() and UnitAffectingCombat("player") then
                self.Globals.CombatTimer:Start()
            elseif self.Globals.CombatTimer:IsRunning() and not UnitAffectingCombat("player") then
                self.Globals.CombatTimer:Reset()
            end

            if self.Enabled then
                self.ObjectManager:Refresh()
                self.UnitManager:ResetObjects()
                self.Globals.MissileManager:Refresh()
                self.TimeToDie:Refresh()
                self:TickModules()
            else
                --Bastion.UnitManager:ResetObjects()
            end
        end
    ]]

    CreateFrame("Frame"):SetScript("OnUpdate", function(this, elapsed)
        self.TimeSinceLastUpdate = self.TimeSinceLastUpdate + elapsed

        if self.TimeSinceLastUpdate >= self.Interval and not self.Updating then
            self.Updating = true
            self.TimeSinceLastUpdate = 0

            if not self.Globals.CombatTimer:IsRunning() and UnitAffectingCombat("player") then
                self.Globals.CombatTimer:Start()
            elseif self.Globals.CombatTimer:IsRunning() and not UnitAffectingCombat("player") then
                self.Globals.CombatTimer:Reset()
            end

            self.ObjectManager:Refresh()
            self.Globals.UnitManager:ResetObjects()
            self.Globals.MissileManager:Refresh()
            self.TimeToDie:Refresh()
            self:TickModules()
            self.Updating = false
        end
    end)

    --Bastion.Globals.EventManager:SetOnUpdate(BastionOnUpdate)

    --[[ Bastion.Ticker = C_Timer.NewTicker(Bastion.Interval, function()
        Bastion.TickFunction(self)
    end) ]]

    Bastion.Globals.Command:Register("dumpspells", "Dump spells to a file", function()
        local i = 1
        local rand = math.random(100000, 999999)
        while true do
            local spellName, spellSubName = C_SpellBook.GetSpellBookItemName(i, BOOKTYPE_SPELL)
            if not spellName then
                do
                    break
                end
            end

            -- use spellName and spellSubName here
            local spellInfo = C_Spell.GetSpellInfo(spellName)

            if spellInfo and spellInfo.spellID then
                spellName = spellName:gsub("[%W%s]", "")
                WriteFile("bastion-" .. UnitClass("player") .. "-" .. rand .. ".lua",
                    "local " .. spellName .. " = Bastion.Globals.SpellBook:GetSpell(" .. spellInfo.spellID .. ")\n", true)
            end
            i = i + 1
        end
    end)

    Bastion.Globals.Command:Register("missed", "Dump the list of immune kidney shot spells", function()
        for k, v in pairs(missed) do
            self.Util:Print(k)
        end
    end)

    self:Require("@/Libraries/")
    self:Require("@/Modules/")
    self:Require("@/")
    LoadThird()
    loaded = true

    self.Loaded = GetTime()
    return self
end

return Bastion
